/***
 * Julia
 * Copyright (C) 2005-2006 INRIA, France Telecom, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliak;

import java.util.HashMap;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.factory.Factory;
import org.objectweb.fractal.api.factory.GenericFactory;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.api.type.TypeFactory;
import org.objectweb.fractal.julia.loader.Loader;
import org.objectweb.fractal.julia.type.BasicComponentType;
import org.objectweb.fractal.julia.type.BasicInterfaceType;
import org.objectweb.fractal.juliak.factory.BasicGenericFactoryImpl;
import org.objectweb.fractal.juliak.factory.TypeFactoryImpl;


/**
 * The Fractal provider class for managing component-based control membranes.
 * This class provides the {@link #newFcInstance()} method for retrieving an
 * instance of the Juliak bootstrap component.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 * @since 2.5
 */
public class Juliak implements Factory {
    
    public Juliak() {}
    
    /**
     * Return an instance of the Fractal provider class.
     * This method is meant to allow subclassing Juliak.
     * This method is thus more general than the default public constructor for
     * obtaining an instance of the provider class.
     * This method returns a singleton instance.
     */
    public static Juliak get() {
        
        if( juliak != null ) {
            return juliak;
        }

        String clname = System.getProperty("fractal.provider");
        if (clname == null) {
            String msg = "The fractal.provider system property should be defined and non empty";
            throw new RuntimeException(msg);
        }

        try {
            Class cl = Class.forName(clname);
            Object o = cl.newInstance();
            juliak = (Juliak) o;
            return juliak;
        }
        catch (Exception e) {
            String msg =
                "Cannot find or instantiate the '" + clname +
                "' class specified in the fractal.provider system property";
            throw new RuntimeException(msg);
        }
    }
    
    /** The singleton instance of the provider class. */
    private static Juliak juliak = null;
    
    /**
     * Property name for dumping run-time generated code.
     * 
     * The value associated to this property contains the directory into which
     * run-time generated code such as Fractal interface implementations are
     * dumped. No dumping if the value is null or empty. Use "." to dump in the
     * current directory.
     * 
     * This property is used in org.objectweb.fractal.juliak.asm.
     */
    final public static String DUMP_GENERATED_PROP_NAME =
        "juliak.loader.gen.dir";
    
    /**
     * Property name for registering additional membrane definitions.
     * 
     * The value associated to this property is a comma-separated list of class
     * names implementing
     * {@link org.objectweb.fractal.juliak.membrane.MembraneDefProvider}.
     * 
     * This property is used by
     * {@link org.objectweb.fractal.juliak.membrane.JuliakMembranes}.
     */
    final public static String CONFIG_PROP_NAME = "juliak.config";
    
    
    /** Return the Fractal type of this component. */
    public Type getFcInstanceType() {
        if( TYPE == null ) {
            try {
                TYPE =
                    new BasicComponentType(
                        new InterfaceType[] {
                            new BasicInterfaceType(
                                "type-factory",
                                TypeFactory.class.getName(),
                                false, false, false),
                            new BasicInterfaceType(
                                "generic-factory",
                                GenericFactory.class.getName(),
                                false, false, false),
                            new BasicInterfaceType(
                                "loader",
                                Loader.class.getName(),
                                false, false, false)
                        });
            }
            catch (InstantiationException e) {
                throw new RuntimeException(e.getMessage());
            }
        }
        return TYPE;
    }

    /** Return the Fractal controller description of this component. */
    public Object getFcControllerDesc() {
        return CONTROLLER_DESC;
    }

    /** Return the Fractal content description of this component. */
    public Object getFcContentDesc() {
        return BootstrapComponentImpl.class.getName();
    }

    /** The singleton instance of the bootstrap component. */
    private static Component bootstrap;
    
    /**
     * Return the reference of the bootstrap component.
     * This method is called by {@link
     * org.objectweb.fractal.api.Fractal#getBootstrapComponent()}
     * The bootstrap component provides the following interfaces:
     * <ul>
     * <li> {@link TypeFactory} </li>
     * <li> {@link GenericFactory} </li>
     * </ul>
     */
    public Component newFcInstance() throws InstantiationException {
        
        if( bootstrap != null )
            return bootstrap;
                
        // Create the bootstrap component
        GenericFactory gf = getFcGenericFactory();
        bootstrap =
            gf.newFcInstance(
                    getFcInstanceType(),
                    getFcControllerDesc(),
                    getFcContentDesc() );
        
        return bootstrap;
    }

    private static ComponentType TYPE = null;    
    final private static Object CONTROLLER_DESC = "bootstrap";

    
    // ----------------------------------------------------------
    // Implementation specific
    // ----------------------------------------------------------
    
    private static TypeFactory tf;
    private static GenericFactory gf;
    private static Loader loader;
    
    public TypeFactory getFcTypeFactory() {
        if( tf == null ) {
            tf = new TypeFactoryImpl();
        }
        return tf;
    }
    
    public GenericFactory getFcGenericFactory() {
        if( gf == null ) {
            gf = new BasicGenericFactoryImpl();
        }
        return gf;
    }
    
    public Loader getFcLoader() {
        
        if( loader != null ) {
            return loader;
        }

        String clname = System.getProperty("julia.loader");
        if( clname==null || clname.length()==0 ) {
            String msg = "The julia.loader system property should be defined and non empty";
            throw new RuntimeException(msg);
        }

        try {
            Class cl = Class.forName(clname);
            Object o = cl.newInstance();
            loader = (Loader) o;
            loader.init(new HashMap());
            return loader;
        }
        catch (Exception e) {
            String msg =
                "Cannot find or instantiate the '" + clname +
                "' class specified in the fractal.provider system property";
            throw new RuntimeException(msg);
        }
    }
    
 }
