/***
 * Julia
 * Copyright (C) 2005-2006 INRIA, France Telecom, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliak.factory;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.factory.GenericFactory;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.loader.Loader;
import org.objectweb.fractal.julia.loader.Tree;
import org.objectweb.fractal.juliak.FractalHelper;
import org.objectweb.fractal.juliak.platform.Platform;


/**
 * Abstract super class for generic factories.
 * 
 * This factory does not make any assumption on the way control membranes
 * are implemented. This concern is deferred to subclasses.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 * @since 2.5
 */
public abstract class AbstractGenericFactoryImpl implements GenericFactory {
    
    // -----------------------------------------------------------------
    // Implementation of the GenericFactory interface
    // -----------------------------------------------------------------
    
    /**
     * Create a new component.
     * 
     * @param type  the component type
     * @param controllerDesc
     *          the description of the control membrane associated
     *          to the component
     * @param contentDesc
     *          the description of the content of the component
     */
    public Component newFcInstance(
            Type type, Object controllerDesc, Object contentDesc )
    throws InstantiationException {
        
        if( controllerDesc instanceof Component ) {
            Component membrane = (Component) controllerDesc;
            return newFcInstance(type, membrane, contentDesc);
        }
        
        /*
         * When called from Fractal ADL, the component creation process may be
         * accompagnied with some hints. One of the them is the class loader to
         * be used. In this case, the controllerDesc parameter is an array where
         * the 1st element contains the class loader.
         */
        if( controllerDesc.getClass().isArray() ) {
            Object[] array = (Object[]) controllerDesc;
            Platform.get().setLoader( array[0] );
            controllerDesc = array[1];
        }
        
        checkFcType(type,controllerDesc);
        checkFcControllerDesc(controllerDesc);
        checkFcContentDesc(type,controllerDesc,contentDesc);
                
        /*
         * Create the content.
         */
        Object content = newFcContent(type,controllerDesc,contentDesc);
        
        /*
         * Create the membrane and bind it to the content.
         */
        Component compctrlitf =
            newFcMembrane(type,controllerDesc,contentDesc,content);
        
        return compctrlitf;
    }

    
    // -----------------------------------------------------------------
    // Membrane related methods
    // -----------------------------------------------------------------
        
    /**
     * Create a new control membrane.
     * 
     * @param type  the type of the component associated to this membrane
     * @param controllerDesc
     *          the controller description of this membrane
     *          (typically a String such as "primitive" or any of the other
     *          registered membrane identifier)
     * @param contentDesc 
     *          the content description of the component associated
     *          to this membrane (e.g. the fully qualified name of the class
     *          implementing a primitive component)
     * @param content
     *          the instance implementing the component associated
     *          to this membrane
     *          
     * @return  the fcinterface for the component interface
     *          exported by the given control membrane  
     */
    abstract protected Component newFcMembrane(
            Type type, Object controllerDesc, Object contentDesc,
            Object content )
    throws InstantiationException;

    /**
     * Bind a control membrane to the component defined by the given type,
     * content description and content.
     * 
     * @param type  the type of the component associated to this membrane
     * @param contentDesc 
     *          the content description of the component associated
     *          to this membrane (e.g. the fully qualified name of the class
     *          implementing a primitive component)
     * @param content
     *          the instance implementing the component associated
     *          to this membrane
     * @param membrane  the control membrane
     * 
     * @return  the fcinterface for the component interface
     *          exported by the given control membrane  
     */
    abstract protected Component bindFcMembrane(
            Type type, Object contentDesc, Object content, Component membrane )
    throws InstantiationException;

        
    // -----------------------------------------------------------------
    // Implementation specific
    // -----------------------------------------------------------------
        
    /**
     * Create a new component.
     * 
     * @param type         the component type
     * @param membrane     the control membrane
     * @param contentDesc  the description of the content of the component
     */
    private Component newFcInstance(
            Type type, Component membrane, Object contentDesc )
    throws InstantiationException {
        
        /*
         * Instantiate the content.
         * contentDesc may be null for composite components.
         */
        Object content = null;
        if( contentDesc != null ) {
            if( !(contentDesc instanceof String) ) {
                final String msg = "contentDesc should be a String";
                throw new InstantiationException(msg);
            }
            String contentDescStr = (String) contentDesc;
            Class cl = loadClass(contentDescStr);
            checkFcContentClassImplementsServerInterfaces(type,cl);
            content = instanciateClass(cl);
        }
        
        /*
         * Bind the membrane to the content.
         */
        Component compctrlitf =
            bindFcMembrane(type,contentDesc,content,membrane);

        return compctrlitf;
    }
        
    
    /**
     * Instantiate the content part of a component.
     */
    protected Object newFcContent(
            Type type, Object controllerDesc, Object contentDesc)
    throws InstantiationException {
        
        /*
         * Adaptation for templates components.
         * Templates are created with something like:
         * Component cTmpl = cf.newFcInstance(
         *      cType, "flatPrimitiveTemplate",
         *      new Object[] { "flatPrimitive", "ClientImpl" });
         */
        if( controllerDesc instanceof String ) {
            if( ((String)controllerDesc).endsWith("Template") ) {
                Object[] cont = (Object[]) contentDesc;
                contentDesc = cont[1];
            }
        }
        
        if( contentDesc == null ) {
            /*
             * contentDesc is null for composite components (this is not
             * mandatory, a composite can have a content, but this is the case
             * most of the time).
             */
            return null;
        }
        
        if( ! (contentDesc instanceof String) ) {
            /*
             * When contentDesc is not a String, Julia assumes that this is not
             * a content descriptor, but that this is the content.
             */
            return contentDesc;
        }
        
        /*
         * Load the content class.
         */
        String contentDescStr = (String) contentDesc;        
        Class contentClass = loadFcContentClass(controllerDesc,contentDescStr);
        
        /*
         * Perform checkings on the content class.
         */
        checkFcContentClassImplementsServerInterfaces(type,contentClass);
        if( isFcContentToBeCheckedForBC(controllerDesc) ) {
            checkFcContentClassforBC(type,contentClass);
        }

        /*
         * Instantiate the content class.
         */
        Object content = null;
        try {
            content = contentClass.newInstance();
        }
        catch(Exception e) {
            final String msg =
                e.getClass().getName()+" when instantiating content class "+
                contentDescStr+". Message: "+e.getMessage();
            throw new InstantiationException(msg);
        }
        
        return content;
    }
    
    /**
     * Load the specified content class.
     * First, try to load the class from the classpath.
     * If that fails, assume that the specified contentDesc corresponds to an
     * alias defined in the julia.cfg configuration file. This is case for
     * control component where the content class is dynamically generated by
     * the ASM mixin class generator.
     * 
     * @param controllerDesc
     *      the controller descriptor of the component for which the content
     *      class is requested
     * @param contentDescStr
     *      the content descriptor. May be a fully-qualified class name or the
     *      stringified representation of a {@link Tree} to be loaded by a Julia
     *      {@link Loader}. 
     */
    protected Class loadFcContentClass( Object controllerDesc, String contentDescStr )
    throws InstantiationException {

        try {
            Class contentClass = Platform.get().loadClass(contentDescStr);
            return contentClass;
        }
        catch(ClassNotFoundException cnfe) {            
            final String msg = "Content class "+contentDescStr+" not found";
            throw new InstantiationException(msg);
        }
    }
    
    
    // -----------------------------------------------------------------
    // Checks
    // -----------------------------------------------------------------
    
    /**
     * Check that the value given for the type is legal.
     */
    protected void checkFcType( Type type, Object controllerDesc )
        throws InstantiationException {
        
        /** type must be a component type. */
        if( ! (type instanceof ComponentType) )
            throw new InstantiationException(
                    "Argument type must be an instance of ComponentType");
        
        /** The Java interfaces referenced in the type must exist. */
        ComponentType ct = (ComponentType) type;
        InterfaceType[] its = ct.getFcInterfaceTypes();
        for (int i = 0; i < its.length; i++) {
            String itSignature = its[i].getFcItfSignature();
            try {
                Platform.get().loadClass(itSignature);
            }
            catch(ClassNotFoundException cnfe) {
                throw new InstantiationException(
                        "Interface "+itSignature+
                        " referenced in Fractal interface type "+
                        its[i].getFcItfName()+" not found");
            }
        }        
    }
        
    /**
     * Check that the given value is a legal controller description.
     */
    protected void checkFcControllerDesc( Object controllerDesc )
    throws InstantiationException {
        
        if( controllerDesc == null ) {
            throw new InstantiationException(
                    "Argument controllerDesc must be non null");        
        }
    }
    
    /**
     * Check that the value given for the content description is legal.
     */
    protected void checkFcContentDesc(
            Type type, Object controllerDesc, Object contentDesc )
    throws InstantiationException {
        
        /** Check for primitive components. */
        if( controllerDesc.equals("flatPrimitive") ||
            controllerDesc.equals("primitive") ||
            controllerDesc.equals("flatParametricPrimitive") ||
            controllerDesc.equals("parametricPrimitive") ) {
            if( contentDesc==null || !(contentDesc instanceof String) )
                throw new InstantiationException(
                        "Argument contentDesc must be a non null String");        
        }
        
        /** Check for template components. */
        if( controllerDesc instanceof String ) {
            if( ((String)controllerDesc).endsWith("Template") ) {
                
                // Templates are created with something like:
                // Component cTmpl = cf.newFcInstance(
                //      cType, "primitiveTemplate",
                //      new Object[] { "primitive", "ClientImpl" });
                
                if( ! contentDesc.getClass().isArray() ) {
                    throw new InstantiationException(
                            "When instantiating a "+controllerDesc+
                            " contentDesc should be an array");
                }
                
                Object[] tab = (Object[]) contentDesc;
                if( tab.length != 2 ) {
                    throw new InstantiationException(
                            "When instantiating a "+controllerDesc+
                            " contentDesc should be an array of size 2");
                }
                
                checkFcControllerDesc(tab[0]);
                checkFcContentDesc(type,tab[0],tab[1]);
            }
        }
    }
    
    /**
     * Check that the content class implements the server interfaces defined in
     * the type.
     */
    protected void checkFcContentClassImplementsServerInterfaces(
            Type type, Class contentClass )
    throws InstantiationException {
        
        ComponentType ct = (ComponentType) type;
        InterfaceType[] its = ct.getFcInterfaceTypes();
        for (int i = 0; i < its.length; i++) {
            String name = its[i].getFcItfName();
            if( ! its[i].isFcClientItf() &&
                FractalHelper.isFcImplementableInterface(name) ) {
                
                String itSignature = its[i].getFcItfSignature();
                
                // Checked before by checkFcType
                // but we need the class here anyway
                Class itClass;
                try {
                    itClass = Platform.get().loadClass(itSignature);
                }
                catch(ClassNotFoundException cnfe) {
                    throw new InstantiationException(
                            "Interface "+itSignature+
                            " referenced in Fractal interface type "+
                            its[i].getFcItfName()+" not found");
                }
                
                if( ! itClass.isAssignableFrom(contentClass) ) {
                    throw new InstantiationException(
                            "Component content class "+contentClass.getName()+
                            " should implement the server interface "+
                            itSignature);
                }
            }
        }
    }

    /**
     * Return true if the content class associated to the given controller
     * description must implement the BindingController interface when its type
     * defines at least one client interface. 
     */
    protected boolean isFcContentToBeCheckedForBC( Object controllerDesc ) {

        /*
         * The content class of component controllers need not implement
         * BindingController.
         */
        if( controllerDesc.equals("mPrimitive") ||
            controllerDesc.equals("mComposite") ) {
            return false;
        }
        
        return true;
    }
    
    /**
     * Check that the content class implements the BindingController interface
     * if at least one client interface is defined in its type.
     */
    protected void checkFcContentClassforBC( Type type, Class contentClass )
    throws InstantiationException {
        
        ComponentType ct = (ComponentType) type;
        InterfaceType[] its = ct.getFcInterfaceTypes();
        for (int i = 0; i < its.length; i++) {
            if( its[i].isFcClientItf() ) {
                if( ! BindingController.class.isAssignableFrom(contentClass) ) {
                    throw new InstantiationException(
                            "Content class "+contentClass.getName()+
                            " is associated to the client interface type "+
                            its[i].getFcItfName()+
                            " and must then implement the "+
                            BindingController.class.getName()+" interface");
                }
            }
        }
    }


    // -----------------------------------------------------------------
    // Utility methods
    // -----------------------------------------------------------------
    
    /**
     * Load and instantiate the given class name.
     */
    private Class loadClass( String name ) throws InstantiationException {
        
        try {
            Class cl = Platform.get().loadClass(name);
            return cl;
        }
        catch( ClassNotFoundException cnfe ) {
            final String msg =
                "ClassNotFoundException when instantiating class: "+name;
            throw new InstantiationException(msg);
        }
    }
    
    /**
     * Load and instantiate the given class name.
     */
    private Object instanciateClass( Class cl ) throws InstantiationException {
        
        try {
            Object o = cl.newInstance();
            return o;
        }
        catch( java.lang.InstantiationException ie ) {
            final String msg =
                "java.lang.InstantiationException when instantiating class: "+
                cl.getName();
            throw new InstantiationException(msg);
        }
        catch( IllegalAccessException iae ) {
            final String msg =
                "IllegalAccessException when instantiating class: "+
                cl.getName();
            throw new InstantiationException(msg);
        }
    }
    
}
