/***
 * Julia
 * Copyright (C) 2005-2006 INRIA, France Telecom, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Lionel.Seinturier@lifl.fr
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliak.factory;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.factory.GenericFactory;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.julia.Julia;
import org.objectweb.fractal.juliak.platform.Platform;
import org.objectweb.fractal.util.Fractal;


/**
 * Abstract super class for generic factories.
 * 
 * This factory does not make any assumption on the way control membranes
 * are implemented. This concern is deferred to subclasses.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@lifl.fr>
 * @since 2.5
 */
public abstract class AbstractJuliaEnabledGenericFactoryImpl
    extends AbstractGenericFactoryImpl {
    
    /**
     * The reference towards the Julia generic factory.
     * Lazily created whenever an heterogeneous assembly containing Julia and
     * Juliak components is instantiated.
     *
     * To obtain a Julia component from the Juliak factory, the convention is
     * to insert the /julia/ prefix in front of the controller description
     * (e.g. /julia/primitive).
     * 
     * This factory differs from {@link AbstractGenericFactoryImpl} by the fact
     * that the instantiation of component with a controller descriptor prefixed
     * with "/julia/" is delegated to the {@link Julia} provider class (these
     * components are then equipped with an OO control membrane).
     */
    private GenericFactory juliagf = null;

    
    // -----------------------------------------------------------------
    // Implementation of the GenericFactory interface
    // -----------------------------------------------------------------
    
    /**
     * Create a new component.
     * 
     * @param type  the component type
     * @param controllerDesc
     *          the description of the control membrane associated
     *          to the component
     * @param contentDesc
     *          the description of the content of the component
     */
    public Component newFcInstance(
            Type type, Object controllerDesc, Object contentDesc )
    throws InstantiationException {
        
        /*
         * When called from Fractal ADL, the component creation process may be
         * accompagnied with some hints. One of the them is the class loader to
         * be used. In this case, the controllerDesc parameter is an array where
         * the 1st element contains the class loader.
         */
        if( controllerDesc.getClass().isArray() ) {
            Object[] array = (Object[]) controllerDesc;
            Platform.get().setLoader( array[0] );
            controllerDesc = array[1];
        }
        
        /*
         * Naming convention: controller descriptions starting with /julia/
         * correspond to components which must be instantiated with Julia.
         */
        if( controllerDesc instanceof String ) {
            String str = (String) controllerDesc;
            if( str.startsWith("/julia/") ) {
                String cdesc = str.substring(7);
                
                // Instantiate the Julia generic factory
                if( juliagf == null ) {
                    Component boot = new Julia().newFcInstance();
                    try {
                        juliagf = Fractal.getGenericFactory(boot);
                    }
                    catch (NoSuchInterfaceException e) {
                        final String msg =
                            "Julia generic factory returned a NoSuchInterfaceException for interface: " +
                            e.getMessage();
                        throw new InstantiationException(msg);
                    }
                }
                
                // Use the Julia generic factory to instantiate the component
                Component comp = juliagf.newFcInstance(type, cdesc, contentDesc);
                return comp;
            }
        }
        
        Component c = super.newFcInstance(type, controllerDesc, contentDesc);
        return c;
    }    
}
